"""Handshake protocol."""

from typing import NamedTuple

from mcipc.query.proto.common import MAGIC, random_session_id, Type


__all__ = ['Request', 'Response']


class Request(NamedTuple):
    """A client → server handshake request packet."""

    magic: bytes
    type: Type
    session_id: int

    def __bytes__(self):
        """Converts the packet to bytes."""
        payload = self.magic
        payload += bytes(self.type)
        payload += self.session_id.to_bytes(4, 'big')
        return payload

    @classmethod
    def create(cls, session_id=None):
        """Returns a handshake request packet with a random session ID."""
        if session_id is None:
            session_id = random_session_id()

        return cls(MAGIC, Type.HANDSHAKE, session_id)


class Response(NamedTuple):
    """A server → client handshake response packet."""

    type: bytes
    session_id: int
    challenge_token: int

    @classmethod
    def from_bytes(cls, bytes_):
        """Creates the packet from bytes."""
        type_ = int.from_bytes(bytes_[0:1], 'big')
        session_id = int.from_bytes(bytes_[1:5], 'big')
        challenge_token = bytes_[5:-1].decode()
        return cls(Type(type_), session_id, int(challenge_token))

    def to_json(self):
        """Returns a JSON-ish dict."""
        return {
            'type': self.type.value, 'session_id': self.session_id,
            'challenge_token': self.challenge_token}



"""Information about online players."""

from re import compile  # pylint: disable=W0622
from typing import NamedTuple, Tuple


__all__ = ['Players']


REGEX = compile('.+ (\\d+) .+ (\\d+) .+: (.*)')


class Players(NamedTuple):
    """Online players information."""

    online: int
    max: int
    names: Tuple[str]

    @classmethod
    def from_response(cls, text):
        """Creates the players information from a server response."""
        match = REGEX.fullmatch(text)
        online, max_, names = match.groups()
        names = filter(None, map(lambda name: name.strip(), names.split(', ')))
        return cls(int(online), int(max_), tuple(names))

    def to_json(self):
        """Returns a JSON-ish dict."""
        return {'online': self.online, 'max': self.max, 'names': self.names}



"""Common enumerations."""

from enum import Enum

from mcipc.server.datatypes import VarInt


__all__ = ['State']


class State(Enum):
    """Protocol state."""

    HANDSHAKE = VarInt(0)
    STATUS = VarInt(1)
    LOGIN = VarInt(2)



"""Represents a seed."""

from re import compile  # pylint: disable=W0622


__all__ = ['Seed']


REGEX = compile('.*\\[(-?\\d+)\\]')


class Seed(int):
    """A seed value."""

    @classmethod
    def from_response(cls, text):
        """Returns a seed from a server response."""
        match = REGEX.fullmatch(text)
        return cls(match.group(1))

    def to_json(self):
        """Returns a JSON-ish dict."""
        return self     # No conversion necessary.



"""An interactive RCON shell."""

from argparse import ArgumentParser
from logging import INFO, basicConfig, getLogger
from sys import exit    # pylint: disable=W0622

from mcipc.config import LOG_FORMAT
from mcipc.rcon.config import CONFIG
from mcipc.rcon.console import PS1, rconcmd


__all__ = ['get_args', 'main']


LOGGER = getLogger('rconshell')


def get_args():
    """Parses and returns the CLI arguments."""

    parser = ArgumentParser(description='An interactive RCON shell.')
    parser.add_argument('server', nargs='?', help="the server's name")
    parser.add_argument('-H', '--host', help="the server's host name")
    parser.add_argument('-p', '--port', type=int, help="the server's port")
    parser.add_argument('-P', '--prompt', default=PS1, help='the shell prompt')
    return parser.parse_args()


def main():
    """Runs the RCON shell."""

    args = get_args()
    basicConfig(level=INFO, format=LOG_FORMAT)
    server = args.server

    if server:
        try:
            host, port, passwd = CONFIG.servers[server]
        except KeyError:
            LOGGER.error('No such server: %s.', server)
            exit(4)
    else:
        host = port = passwd = None

    host = args.host or host
    port = args.port or port
    prompt = args.prompt
    exit_code = rconcmd(host, port, passwd, prompt)
    exit(exit_code)



"""More complex data structures."""

from json import dumps, loads
from typing import NamedTuple

from mcipc.server.datatypes import VarInt
from mcipc.server.enumerations import State


__all__ = ['Handshake']


class Handshake(NamedTuple):
    """Server-bound handshake packet."""

    protocol: VarInt
    address: str
    port: int
    next_state: State

    @classmethod
    def from_bytes(cls, bytes_):
        """Creates a handshake object from the respective bytes."""
        version = VarInt.from_bytes(bytes_[1:4])
        address = bytes_[4:-3].decode()
        port = int.from_bytes(bytes_[-3:-1], 'little')
        next_state = VarInt.from_bytes(bytes_[-1:])
        return cls(version, address, port, State(next_state))



class SLPResponse(dict):
    """A server list ping response."""

    def __bytes__(self):
        """Returns the respective bytes."""
        string = dumps(self)
        payload = string.encode('latin-1')
        payload_size = len(payload)
        payload_size = VarInt(payload_size)
        payload_size = bytes(payload_size)
        payload = payload_size + payload
        payload = b'\x00' + payload     # Add packet ID.
        payload_size = len(payload)
        payload_size = VarInt(payload_size)
        payload_size = bytes(payload_size)
        return payload_size + payload

    @classmethod
    def from_bytes(cls, payload):
        """Creates the SLP response from the respective payload."""
        json_bytes = payload[5:]    # Skip total size, packet ID and JSON size.
        string = json_bytes.decode('latin-1')
        return cls(loads(string))



"""Data types for the server protocol."""

from mcipc.server.functions import rshift


__all__ = ['VarInt']


class VarInt(int):
    """Minecraft protocol VarInt type."""

    def __bytes__(self):
        """Returns the respective bytes."""
        if self == 0:
            return self.to_bytes(1, 'little')

        value = int(self)
        bytes_ = b''

        while True:
            temp = value & 0b01111111
            value = rshift(value, 7)

            if value:
                temp |= 0b10000000

            bytes_ += temp.to_bytes(1, 'little')

            if not value:
                break

        return bytes_

    @classmethod
    def from_bytes(cls, bytes_):
        """Reads a VarInt from the respective bytes."""
        if len(bytes_) > 5:
            raise ValueError('VarInt is too big.')

        result = 0

        for index, byte in enumerate(bytes_):
            value = byte & 0b01111111
            shift = 7 * index
            result |= value << shift

            if (byte & 0b10000000) == 0:
                break

        return cls(result)



"""Basic statistics protocol."""

from ipaddress import IPv4Address
from typing import NamedTuple

from mcipc.query.proto.common import MAGIC, random_session_id, Type


__all__ = ['Request', 'BasicStats']


class Request(NamedTuple):
    """Basic statistics request packet."""

    magic: bytes
    type: Type
    session_id: int
    challenge_token: int

    def __bytes__(self):
        """Returns the packet as bytes."""
        payload = self.magic
        payload += bytes(self.type)
        payload += self.session_id.to_bytes(4, 'big')
        payload += self.challenge_token.to_bytes(4, 'big')
        return payload

    @classmethod
    def create(cls, challenge_token, session_id=None):
        """Creates a new request with the specified challenge
        token and the specified or a random session ID.
        """
        if session_id is None:
            session_id = random_session_id()

        return cls(MAGIC, Type.STAT, session_id, challenge_token)


class BasicStats(NamedTuple):
    """Basic statistics response packet."""

    type: Type
    session_id: int
    motd: str
    game_type: str
    map: str
    num_players: int
    max_players: int
    host_port: int
    host_ip: IPv4Address

    @classmethod
    def from_bytes(cls, bytes_):    # pylint: disable=R0914
        """Creates the packet from the respective bytes."""
        type_ = int.from_bytes(bytes_[0:1], 'big')
        session_id = int.from_bytes(bytes_[1:5], 'big')

        try:
            motd, *blocks, port_ip, _ = bytes_[5:].split(b'\0')
        except ValueError:
            raise ValueError('Unexpected amount of Null terminated strings.')

        motd = motd.decode('latin-1')
        strings = [block.decode() for block in blocks]

        try:
            game_type, map_, num_players, max_players = strings
        except ValueError:
            raise ValueError('Unexpected amount of string fields.')

        num_players = int(num_players)
        max_players = int(max_players)
        host_port = int.from_bytes(port_ip[0:2], 'little')
        host_ip = IPv4Address(port_ip[2:].decode())
        return cls(
            Type(type_), session_id, motd, game_type, map_, num_players,
            max_players, host_port, host_ip)

    def to_json(self, ip_type=str):
        """Returns a JSON-ish dict."""
        host_ip = ip_type(self.host_ip)
        return {
            'type': self.type.value, 'session_id': self.session_id,
            'motd': self.motd, 'game_type': self.game_type, 'map': self.map,
            'num_players': self.num_players, 'max_players': self.max_players,
            'host_port': self.host_port, 'host_ip': host_ip}



"""An interactive console."""

from getpass import getpass

from mcipc.rcon.client import Client
from mcipc.rcon.exceptions import InvalidCredentialsError
from mcipc.rcon.exceptions import RequestIdMismatchError


__all__ = ['rconcmd']


PS1 = 'RCON> '
EXIT_COMMANDS = {'exit', 'quit'}
MSG_QUERY_LATER = '\nOkay, I will ask again later.'
MSG_ABORTED = '\nAborted...'
MSG_LOGIN_ABORTED = '\nLogin aborted. Bye.'
MSG_EXIT = 'Bye.'
MSG_SESSION_TIMEOUT = 'Session timed out. Please login again.'
MSG_EXIT_USAGE = 'Usage: {} [<exit_code>].'


def _read(prompt: str, typ=None):
    """Reads input and converts it to the respective type."""

    while True:
        raw = input(prompt)

        if typ is not None:
            try:
                return typ(raw)
            except (TypeError, ValueError):
                print(f'Invalid {typ}: {raw}.')
                continue

        return raw


def _read_or_none(prompt: str, typ=None):
    """Reads the input and returns None on abort."""

    try:
        return _read(prompt, typ=typ)
    except EOFError:
        print(MSG_QUERY_LATER)
        return None


def _login(client: Client, passwd: str):
    """Performs a login."""

    if passwd is None:
        passwd = getpass('Password: ')

    logged_in = False

    while not logged_in:
        try:
            logged_in = client.login(passwd)
        except InvalidCredentialsError:
            print('Invalid password.')
            passwd = getpass('Password: ')

    return passwd


def _read_args(host: str, port: int, passwd: str, prompt: str) -> tuple:
    """Reads the necessary arguments."""

    while any(item is None for item in (host, port, passwd, prompt)):
        if host is None:
            host = _read_or_none('Host: ')

        if port is None:
            port = _read_or_none('Port: ', typ=int)

        if passwd is None:
            passwd = _read_or_none('Password: ')

        if prompt is None:
            prompt = _read_or_none('Prompt: ')

    return (host, port, passwd, prompt)


def _exit(exit_code=0):
    """Exits the interactive shell via exit command."""

    exit_code = int(exit_code)
    print(MSG_EXIT)
    return exit_code


def rconcmd(host: str, port: int, passwd: str, prompt: str = PS1) -> int:
    """Initializes the console."""

    try:
        host, port, passwd, prompt = _read_args(host, port, passwd, prompt)
    except KeyboardInterrupt:
        print(MSG_ABORTED)
        return 1

    with Client(host, port) as client:
        try:
            passwd = _login(client, passwd)
        except (EOFError, KeyboardInterrupt):
            print(MSG_LOGIN_ABORTED)
            return 1

        while True:
            try:
                command = input(prompt)
            except EOFError:
                print(f'\n{MSG_EXIT}')
                break
            except KeyboardInterrupt:
                print()
                continue

            try:
                command, *args = command.split()
            except ValueError:
                continue

            if command in EXIT_COMMANDS:
                try:
                    return _exit(*args)
                except (TypeError, ValueError):
                    print(MSG_EXIT_USAGE.format(command))
                    continue

            try:
                result = client.run(command, *args)
            except RequestIdMismatchError:
                print(MSG_SESSION_TIMEOUT)

                try:
                    passwd = _login(client, passwd)
                except (EOFError, KeyboardInterrupt):
                    print(MSG_LOGIN_ABORTED)
                    return 2

            print(result)

    return 0



"""Locations."""

from re import compile  # pylint: disable=W0622
from typing import NamedTuple


__all__ = ['Location']


REGEX = compile('.*\\[(-?\\d+), (~|-?\\d+), (-?\\d+)\\].*')


def _int_or_none(string):
    """Returns None iff coordinate is special "~"
    character or else the respective integer value.
    """

    if string == '~':
        return None

    return int(string)


class Location(NamedTuple):
    """A 3D location."""

    x: int
    y: int
    z: int

    @classmethod
    def from_response(cls, text):
        """Creates a location from a server response."""
        match = REGEX.fullmatch(text)
        coordinates = (_int_or_none(item) for item in match.groups())
        return cls(*coordinates)

    def to_json(self):
        """Returns a JSON-ish dict."""
        return {'x': self.x, 'y': self.y, 'z': self.z}



"""Low-level protocol stuff."""

from enum import Enum
from logging import getLogger
from random import randint
from typing import NamedTuple

from mcipc.common import BaseClient
from mcipc.rcon.exceptions import InvalidCredentialsError
from mcipc.rcon.exceptions import InvalidPacketStructureError
from mcipc.rcon.exceptions import RequestIdMismatchError


__all__ = ['Type', 'Packet', 'Client']


LOGGER = getLogger(__file__)
TAIL = b'\0\0'


def _rand_uint32() -> int:
    """Returns a random unsigned int32."""

    return randint(0, 4_294_967_295 + 1)


class Type(Enum):
    """Available packet types."""

    LOGIN = 3
    COMMAND = 2
    RESPONSE = 0

    def __bytes__(self):
        """Returns the integer value as little endian."""
        return self.value.to_bytes(4, 'little')     # pylint: disable=E1101


class Packet(NamedTuple):
    """An RCON packet."""

    request_id: int
    type: Type
    payload: str

    def __bytes__(self):
        """Returns the packet as bytes."""
        payload = self.request_id.to_bytes(4, 'little')
        payload += bytes(self.type)
        payload += self.payload.encode()
        payload += TAIL
        size = len(payload).to_bytes(4, 'little')
        return size + payload

    @classmethod
    def from_bytes(cls, bytes_: bytes):
        """Creates a packet from the respective bytes."""
        request_id = int.from_bytes(bytes_[:4], 'little')
        type_ = int.from_bytes(bytes_[4:8], 'little')
        payload = bytes_[8:-2]
        tail = bytes_[-2:]

        if tail != TAIL:
            raise InvalidPacketStructureError('Invalid tail.', tail)

        return cls(request_id, Type(type_), payload.decode())

    @classmethod
    def from_command(cls, command: str):
        """Creates a command packet."""
        return cls(_rand_uint32(), Type.COMMAND, command)

    @classmethod
    def from_login(cls, passwd: str):
        """Creates a login packet."""
        return cls(_rand_uint32(), Type.LOGIN, passwd)


class Client(BaseClient):
    """An RCON client."""

    def communicate(self, packet: Packet) -> Packet:
        """Sends and receives a packet."""
        self._socket.send(bytes(packet))
        header = self._socket.recv(4)
        length = int.from_bytes(header, 'little')
        payload = self._socket.recv(length)
        response = Packet.from_bytes(payload)

        if response.request_id == packet.request_id:
            return response

        raise RequestIdMismatchError(packet.request_id, response.request_id)

    def login(self, passwd: str) -> bool:
        """Performs a login."""
        packet = Packet.from_login(passwd)

        try:
            self.communicate(packet)
        except RequestIdMismatchError:
            raise InvalidCredentialsError()

        return True

    def run(self, command: str, *arguments: str, raw: bool = False) -> str:
        """Runs a command."""
        command = ' '.join((command,) + arguments)
        packet = Packet.from_command(command)
        response = self.communicate(packet)
        return response if raw else response.payload



"""Full statistics protocol."""

from ipaddress import IPv4Address
from typing import NamedTuple

from mcipc.query.proto.common import MAGIC, random_session_id, Type


__all__ = ['Request', 'FullStats']


PADDING = 0
ZERO = b'\0'


def get_dict(bytes_):
    """Returns the end index and a dictionary
    of zero-separated key-value pairs.
    """

    item = b''
    dictionary = {}
    key = None
    is_key = True

    for index, integer in enumerate(bytes_):
        byte = bytes([integer])

        if byte == ZERO:
            if not item and is_key:
                return (index, dictionary)

            if is_key:
                key = item
                is_key = False
            else:
                dictionary[key] = item
                key = None
                is_key = True

            item = b''
        else:
            item += byte

    raise ValueError('Bytes string not properly terminated.', bytes_)


def items(bytes_):
    """Yields zero-byte-separated items."""

    item = b''

    for integer in bytes_:
        byte = bytes([integer])

        if byte == ZERO:
            if not item:
                return

            yield item
            item = b''
        else:
            item += byte


def plugins_to_dict(string):
    """Convers a plugins string into a dictionary."""

    try:
        mod, plugins = string.split(': ')
    except ValueError:  # No plugins.
        return {}

    plugins = plugins.split('; ')
    return {mod: tuple(plugins)}


def stats_from_dict(dictionary):
    """Yields statistics options from the provided dictionary."""

    yield dictionary[b'hostname'].decode('latin-1')
    yield dictionary[b'gametype'].decode()
    yield dictionary[b'game_id'].decode()
    yield dictionary[b'version'].decode()
    yield plugins_to_dict(dictionary[b'plugins'].decode('latin-1'))
    yield dictionary[b'map'].decode()
    yield int(dictionary[b'numplayers'].decode())
    yield int(dictionary[b'maxplayers'].decode())
    yield int(dictionary[b'hostport'].decode())
    yield IPv4Address(dictionary[b'hostip'].decode())


class Request(NamedTuple):
    """Basic statistics request packet."""

    magic: bytes
    type: Type
    session_id: int
    challenge_token: int
    padding: int

    def __bytes__(self):
        """Returns the packet as bytes."""
        payload = self.magic
        payload += bytes(self.type)
        payload += self.session_id.to_bytes(4, 'big')
        payload += self.challenge_token.to_bytes(4, 'big')
        payload += self.padding.to_bytes(4, 'big')
        return payload

    @classmethod
    def create(cls, challenge_token, session_id=None):
        """Creates a new request with the specified challenge
        token and the specified or a random session ID.
        """
        if session_id is None:
            session_id = random_session_id()

        return cls(MAGIC, Type.STAT, session_id, challenge_token, PADDING)


class FullStats(NamedTuple):
    """Full statistics response."""

    type: Type
    session_id: int
    host_name: str
    game_type: str
    game_id: str
    version: str
    plugins: dict
    map: str
    num_players: int
    max_players: int
    host_port: int
    host_ip: IPv4Address
    players: tuple

    @classmethod
    def from_bytes(cls, bytes_):
        """Creates the full stats object from the respective bytes."""
        type_ = int.from_bytes(bytes_[0:1], 'big')
        session_id = int.from_bytes(bytes_[1:5], 'big')
        index = 16  # Discard padding.
        index, stats = get_dict(bytes_[index:])
        index += 16 + 1     # Discard additional null byte.
        index += 10     # Discard padding.
        players = tuple(player.decode() for player in items(bytes_[index:]))
        return cls(Type(type_), session_id, *stats_from_dict(stats), players)

    def to_json(self, ip_type=str):
        """Returns a JSON-ish dict."""
        host_ip = ip_type(self.host_ip)
        return {
            'type': self.type.value, 'session_id': self.session_id,
            'host_name': self.host_name, 'game_type': self.game_type,
            'game_id': self.game_id, 'version': self.version,
            'plugins': self.plugins, 'map': self.map,
            'num_players': self.num_players, 'max_players': self.max_players,
            'host_port': self.host_port, 'host_ip': host_ip,
            'players': self.players}



"""Query client library."""

from socket import SOCK_DGRAM

from mcipc.common import BaseClient
from mcipc.query.proto import basic_stats, full_stats, handshake


__all__ = ['Client']


class Client(BaseClient):
    """A Query client."""

    def __init__(self, host: str, port: int):
        """Sets the challenge token."""
        super().__init__(host, port, typ=SOCK_DGRAM)
        self._challenge_token = None

    def __enter__(self):
        """Performs a handshake."""
        super().__enter__()

        if self._challenge_token is None:
            self._challenge_token = self.handshake().challenge_token

        return self

    def _recv_all(self, buffer: int = 4096):
        """Recevies all bytes."""
        bytes_ = b''

        while True:
            chunk = self._socket.recv(buffer)
            bytes_ += chunk

            if len(chunk) < buffer:
                break

        return bytes_

    def communicate(self, packet, response_type=None, *, buffer: int = 4096):
        """Sends and receives a packet."""
        self._socket.send(bytes(packet))
        response = self._recv_all(buffer=buffer)

        if response_type is not None:
            return response_type.from_bytes(response)

        return response

    def handshake(self):
        """Performs a handshake."""
        request = handshake.Request.create()
        return self.communicate(request, handshake.Response)

    @property
    def basic_stats(self):
        """Returns basic stats"""
        request = basic_stats.Request.create(self._challenge_token)
        return self.communicate(request, basic_stats.BasicStats)

    @property
    def full_stats(self):
        """Returns full stats"""
        request = full_stats.Request.create(self._challenge_token)
        return self.communicate(request, full_stats.FullStats)



"""The actual server."""

from logging import getLogger
from socket import socket

from mcipc.server.datastructures import Handshake, SLPResponse
from mcipc.server.datatypes import VarInt
from mcipc.server.enumerations import State


__all__ = ['StubServer']


LOGGER = getLogger(__file__)


class StubServer:
    """A stub minecraft server."""

    def __init__(self, description, max_players=20, protocol=485):
        """Description, max players and protocol information."""
        self.description = description
        self.max_players = max_players
        self.protocol = protocol

    @property
    def slp_response(self):
        """Returns an SLP response."""
        json = {
            'version': {
                'name': '1.14.2',
                'protocol': self.protocol
            },
            'players': {
                'max': self.max_players,
                'online': 0
            },
            'description': {
                'text': self.description
            }
        }
        return SLPResponse(json)

    @staticmethod
    def _perform_handshake(connection):
        """Handle handshake requests."""
        header = connection.recv(1)
        size = VarInt.from_bytes(header)
        LOGGER.debug('Read size: %s', size)
        payload = connection.recv(size)
        handshake = Handshake.from_bytes(payload)
        LOGGER.debug('Got handshake: %s', handshake)

    def _perform_status(self, connection):
        """Handles status requests."""
        _ = connection.recv(1)  # Discard packet ID.
        response = bytes(self.slp_response)
        connection.send(response)

    def _process(self, connection):
        """Runs the connection processing."""
        state = State.HANDSHAKE

        with connection:
            if state == State.HANDSHAKE:
                LOGGER.debug('HANDSHAKE')
                self._perform_handshake(connection)
                self._perform_status(connection)
                state = State.STATUS
            elif state == State.STATUS:
                LOGGER.debug('STATUS')
                payload = connection.recv(1024)
                LOGGER.debug('Payload: %s', payload)

    def spawn(self, address, port):
        """Spawns the server on the respective socket."""
        with socket() as sock:
            sock.bind((address, port))
            sock.listen()

            while True:
                connection, address = sock.accept()
                LOGGER.debug('New connection from: %s', address)
                self._process(connection)



"""Common protocol stuff."""

from enum import Enum
from random import randint


__all__ = ['MAGIC', 'random_session_id', 'Type']


MAGIC = b'\xfe\xfd'


def random_session_id():
    """Returns a random session ID."""

    return randint(0, 32)


class Type(Enum):
    """Request type."""

    HANDSHAKE = 9
    STAT = 0

    def __bytes__(self):
        """Reuturns the integer value as big endian."""
        return self.value.to_bytes(1, 'big')    # pylint: disable=E1101



"""Common functions for multiple CLI programs."""

from getpass import getpass
from logging import getLogger
from sys import exit    # pylint: disable=W0622

from mcipc.config import InvalidCredentialsError, Credentials
from mcipc.rcon.config import CONFIG


__all__ = ['get_creadentials']


LOGGER = getLogger(__file__)


def get_creadentials(server, logger=LOGGER):
    """Get the credentials for a server from the respective server name."""

    try:
        host, port, passwd = Credentials.from_string(server)
    except InvalidCredentialsError:
        try:
            host, port, passwd = CONFIG.servers[server]
        except KeyError:
            logger.error('No such server: %s.', server)
            exit(2)

    if passwd is None:
        try:
            passwd = getpass('Password: ')
        except (KeyboardInterrupt, EOFError):
            print()
            logger.error('Aborted by user.')
            exit(3)

    return (host, port, passwd)



"""RCON exceptions."""


__all__ = [
    'InvalidPacketStructureError',
    'RequestIdMismatchError',
    'InvalidCredentialsError']


class InvalidPacketStructureError(Exception):
    """Indicates an invalid packet structure."""


class RequestIdMismatchError(Exception):
    """Indicates that the sent and received request IDs do not match."""

    def __init__(self, sent, received):
        """Sets the sent and received request IDs."""
        super().__init__(sent, received)
        self.sent = sent
        self.received = received


class InvalidCredentialsError(Exception):
    """Indicates invalid RCON password."""



"""RCON client CLI."""

from argparse import ArgumentParser
from logging import DEBUG, INFO, basicConfig, getLogger
from subprocess import CalledProcessError, check_call
from sys import exit    # pylint: disable=W0622

from mcipc.cli.common import get_creadentials
from mcipc.config import LOG_FORMAT
from mcipc.rcon.playground import Client


__all__ = ['get_args', 'run_action', 'main']


LOGGER = getLogger('rconclt')


def get_args():
    """Parses and returns the CLI arguments."""

    parser = ArgumentParser(description='A Minecraft RCON client.')
    parser.add_argument('server', help="the server's name")
    subparsers = parser.add_subparsers(dest='action')
    command_parser = subparsers.add_parser(
        'exec', help='execute commands on the server')
    command_parser.add_argument(
        'command', help='command to execute on the server')
    command_parser.add_argument(
        'argument', nargs='*', default=(), help='arguments for the command')
    say_parser = subparsers.add_parser(
        'say', help='broadcast a message on the server')
    say_parser.add_argument('message', help='the message to broadcast')
    fortune_parser = subparsers.add_parser(
        'fortune', help='send a fortune to the players on the server')
    fortune_parser.add_argument(
        '-l', '--long', action='store_true', help='generate ling fortunes')
    fortune_parser.add_argument(
        '-o', '--offensive', action='store_true',
        help='generate offensive fortunes')
    datetime_parser = subparsers.add_parser(
        'datetime',
        help='sends the current date and time to the players on the server')
    datetime_parser.add_argument(
        '-f', '--format', default='%c', help='the datetime format')
    subparsers.add_parser('in-use', help='checks whether the server is in use')
    shutdown_parser = subparsers.add_parser(
        'idle-shutdown', help='shuts down the server if it is not in use')
    shutdown_parser.add_argument(
        '-s', '--sudo', action='store_true',
        help='invoke the shutdown command using sudo')
    shutdown_parser.add_argument(
        '-u', '--unit', default='minecraft@{server}.service',
        help='the systemd unit template')
    parser.add_argument(
        '-d', '--debug', action='store_true',
        help='print additional debug information')
    return parser.parse_args()


def idle_shutdown(players, args):
    """Shuts down the server if it is idle."""

    if players.online:
        LOGGER.info('Server is in use.')
        return False

    LOGGER.info('Server is idle.')
    unit = args.unit.format(server=args.server)
    command = ('/usr/bin/systemctl', 'stop', unit)

    try:
        check_call(command)
    except CalledProcessError as error:
        LOGGER.error('Could not shutdown the server.')
        LOGGER.debug(error)
        return False

    LOGGER.info('Server %s has been shut down.', unit)
    return True


def run_action(client, args):
    """Runs the respective actions."""

    result = None

    if args.action == 'exec':
        result = client.run(args.command, *args.argument)
    elif args.action == 'say':
        result = client.say(args.message)
    elif args.action == 'fortune':
        result = client.fortune(
            short=not args.long, offensive=args.offensive)
    elif args.action == 'datetime':
        result = client.datetime(frmt=args.format)
    elif args.action == 'in-use':
        players = client.players

        if players.online:
            LOGGER.info('There are %i players online:', players.online)
            LOGGER.info(', '.join(players.names))
        else:
            LOGGER.warning('There are no players online.')
            exit(1)

    if result:
        LOGGER.info(result)


def main():
    """Runs the RCON client."""

    args = get_args()
    log_level = DEBUG if args.debug else INFO
    basicConfig(level=log_level, format=LOG_FORMAT)
    host, port, passwd = get_creadentials(args.server, logger=LOGGER)

    with Client(host, port) as client:
        if not client.login(passwd):
            LOGGER.error('Failed to log in.')
            exit(4)

        if args.action == 'idle-shutdown':
            players = client.players
        else:
            run_action(client, args)

    if args.action == 'idle-shutdown':
        if not idle_shutdown(players, args):
            exit(1)
